﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using TownGenerator.TownGeneration;

namespace TownGenerator
{
    public sealed class Town
    {
        private Tile[,] tiles;
        private List<Node> nodes;

        public Town(Tile[,] tiles, List<Node> nodes)
        {
            this.tiles = tiles;
            this.nodes = nodes;

            Bitmap = CreateBitmap();
        }

        public Bitmap Bitmap { get; }

        public List<Node> GetNodes() { return nodes; }

        public List<HashSet<Edge>> FindEdges()
        {
            var listOfEdgesBetweenEveryNode = new List<HashSet<Edge>>();

            // From every node find all edges to all other nodes
            // Eliminates duplicates
            for (int i = 0; i < nodes.Count - 1; i++)
            {
                for (int j = i + 1; j < nodes.Count; j++)
                {
                    if (!nodes[i].Equals(nodes[j]))
                    {
                        Crossroad origin = Crossroad.FromNode(nodes[i], tiles);
                        HashSet<Edge> edges = new HashSet<Edge>();
                        FindEdges(edges, origin, nodes[j], 0, nodes[i], origin);
                        listOfEdgesBetweenEveryNode.Add(edges);
                    }
                }
            }

            return listOfEdgesBetweenEveryNode;
        }

        public void FindEdges(HashSet<Edge> edges, Node origin, Node target, int distance, Node currentNode, Crossroad currentCrossroad)
        {
            int x = currentNode.X;
            int y = currentNode.Y;
            
            if (tiles[x, y].TileType == TileType.Building)
                return;

            if (x == target.X && y == target.Y) // Pokud dojdu do hledaného uzlu
            {
                edges.Add(new Edge(origin, target, distance));
                return;
            }

            Direction direction;
            while ((direction = currentCrossroad.GetDirection()) != Direction.None)
            {
                switch (direction)
                {
                    case Direction.Right:
                        {
                            currentCrossroad.CanGoRight = false;
                            Crossroad next = NextCrossroad(edges, origin, target, distance + 1, new Node(x + 1, y), currentCrossroad, Direction.Right);
                            if (next == null)
                                continue;
                            else
                            {
                                if (!HasAlreadyGoneThrough(next))
                                    FindEdges(edges, origin, target, next.Distance, next, next);
                            }
                        }
                        break;
                    case Direction.Up:
                        {
                            currentCrossroad.CanGoUp = false;
                            Crossroad next = NextCrossroad(edges, origin, target, distance + 1, new Node(x, y - 1), currentCrossroad, Direction.Up);
                            if (next == null)
                                continue;
                            else
                            {
                                if (!HasAlreadyGoneThrough(next))
                                    FindEdges(edges, origin, target, next.Distance, next, next);
                            }
                        }
                        break;
                    case Direction.Left:
                        {
                            currentCrossroad.CanGoLeft = false;
                            Crossroad next = NextCrossroad(edges, origin, target, distance + 1, new Node(x - 1, y), currentCrossroad, Direction.Left);
                            if (next == null)
                                continue;
                            else
                            {
                                if (!HasAlreadyGoneThrough(next))
                                    FindEdges(edges, origin, target, next.Distance, next, next);
                            }
                        }
                        break;
                    case Direction.Down:
                        {
                            currentCrossroad.CanGoDown = false;
                            Crossroad next = NextCrossroad(edges, origin, target, distance + 1, new Node(x, y + 1), currentCrossroad, Direction.Down);
                            if (next == null)
                                continue;
                            else
                            {
                                if (!HasAlreadyGoneThrough(next))
                                    FindEdges(edges, origin, target, next.Distance, next, next);
                            }
                        }
                        break;
                }
            }
            return;
        }

        public Crossroad NextCrossroad(HashSet<Edge> edges, Node origin, Node target, int distance, Node currentNode,
            Crossroad lastCrossroad, Direction lastDirection)
        {
            int x = currentNode.X;
            int y = currentNode.Y;

            if (x == target.X && y == target.Y) // Pokud dojdu do hledaného uzlu
            {
                edges.Add(new Edge(origin, target, distance));
                return null;
            }

            Crossroad nextCrossroad;
            if (lastCrossroad.X != currentNode.X || lastCrossroad.Y != currentNode.Y) // nechceme, aby se nám vracela ta samá křižovatka, z které jsme přišli
                if (Crossroad.TryGetCrossroad(currentNode, tiles, lastCrossroad, lastDirection, distance, out nextCrossroad))
                    return nextCrossroad;

            // Turn right
            if (lastDirection != Direction.Left)
            {
                if (x + 1 < tiles.GetLength(0))
                {
                    if (tiles[x + 1, y].TileType == TileType.Road)
                        return NextCrossroad(edges, origin, target, distance + 1, new Node(x + 1, y), lastCrossroad, Direction.Right);
                }
            }
            // Turn up
            if (lastDirection != Direction.Down)
            {
                if (y - 1 >= 0)
                {
                    if (tiles[x, y - 1].TileType == TileType.Road)
                        return NextCrossroad(edges, origin, target, distance + 1, new Node(x, y - 1), lastCrossroad, Direction.Up);
                }
            }
            // Turn left
            if (lastDirection != Direction.Right)
            {
                if (x - 1 >= 0)
                {
                    if (tiles[x - 1, y].TileType == TileType.Road)
                        return NextCrossroad(edges, origin, target, distance + 1, new Node(x - 1, y), lastCrossroad, Direction.Left);
                }
            }
            // Turn down
            if (lastDirection != Direction.Up)
            {
                if (y + 1 < tiles.GetLength(1))
                {
                    if (tiles[x, y + 1].TileType == TileType.Road)
                        return NextCrossroad(edges, origin, target, distance + 1, new Node(x, y + 1), lastCrossroad, Direction.Down);
                }
            }
            return null;
        }

        private Bitmap CreateBitmap()
        {
            int tileWidth = 32;
            Image image = Properties.Resources.Tileset;
            Bitmap result = new Bitmap(tileWidth * tiles.GetLength(0), tileWidth * tiles.GetLength(1));

            using (Graphics g = Graphics.FromImage(result))
            {
                for (int y = 0; y < tiles.GetLength(1); y++)
                {
                    for (int x = 0; x < tiles.GetLength(0); x++)
                    {
                        using (TextureBrush brush = new TextureBrush(image, WrapMode.Tile, GetTileRect(tiles[x, y].GetTileIndex())))
                        {
                            g.FillRectangle(brush, x * tileWidth, y * tileWidth, tileWidth, tileWidth);
                        }
                    }
                }
            }
            return result;
        }
        
        private Rectangle GetTileRect(int index)
        {
            int x = 0, y = 0;
            int size = 32;

            if (1 <= index && index <= 6)
            {
                y = 0;
                x = (index - 1) * size;
            }
            else if (7 <= index && index <= 11)
            {
                y = size;
                x = (index - 7) * size;
            }
            else if (12 <= index && index <= 16)
            {
                y = 2 * size;
                x = (index - 12) * size;
            }
            return new Rectangle(x, y, size, size);
        }

        private bool HasAlreadyGoneThrough(Crossroad crossroad)
        {
            Crossroad old = crossroad.LastCrossroad;
            if (old.X == crossroad.X && old.Y == crossroad.Y)
                return true;

            while ((old = old.LastCrossroad) != null)
            {
                if (old.X == crossroad.X && old.Y == crossroad.Y)
                    return true;
            }
            return false;
        }
        
    }
}

